@@ -35,6 +35,80 @@ class ServerState(Enum):
     LOCKED = "locked"        # Bound to device, rejecting others
 
 
+class ChatHistory:
+    """Persistent chat history storage."""
+
+    def __init__(self, sessions_dir: Path):
+        self.sessions_dir = sessions_dir
+        self.sessions_dir.mkdir(parents=True, exist_ok=True)
+
+    def _get_session_file(self, session_id: str) -> Path:
+        """Get the file path for a session's chat history."""
+        # Use a hash of the session_id for the filename
+        safe_name = hashlib.sha256(session_id.encode()).hexdigest()[:16]
+        return self.sessions_dir / f"{safe_name}.json"
+
+    def load(self, session_id: str) -> list[dict]:
+        """Load chat history for a session."""
+        file_path = self._get_session_file(session_id)
+        if file_path.exists():
+            try:
+                with open(file_path, "r") as f:
+                    data = json.load(f)
+                    return data.get("messages", [])
+            except (json.JSONDecodeError, IOError):
+                return []
+        return []
+
+    def save_message(self, session_id: str, message: dict):
+        """Append a message to the chat history."""
+        file_path = self._get_session_file(session_id)
+
+        # Load existing messages
+        messages = self.load(session_id)
+        messages.append({
+            **message,
+            "timestamp": time.time()
+        })
+
+        # Keep only last 1000 messages per session
+        if len(messages) > 1000:
+            messages = messages[-1000:]
+
+        # Save
+        try:
+            with open(file_path, "w") as f:
+                json.dump({"session_id": session_id, "messages": messages}, f)
+        except IOError:
+            pass  # Ignore save errors
+
+    def list_sessions(self) -> list[dict]:
+        """List all saved sessions with metadata."""
+        sessions = []
+        for file_path in self.sessions_dir.glob("*.json"):
+            try:
+                with open(file_path, "r") as f:
+                    data = json.load(f)
+                    session_id = data.get("session_id", "")
+                    messages = data.get("messages", [])
+                    if messages:
+                        last_msg = messages[-1]
+                        sessions.append({
+                            "session_id": session_id,
+                            "message_count": len(messages),
+                            "last_active": last_msg.get("timestamp", 0),
+                            "preview": last_msg.get("content", "")[:100] if "content" in last_msg else ""
+                        })
+            except (json.JSONDecodeError, IOError):
+                continue
+        return sorted(sessions, key=lambda x: x.get("last_active", 0), reverse=True)
+
+    def delete(self, session_id: str):
+        """Delete a session's chat history."""
+        file_path = self._get_session_file(session_id)
+        file_path.unlink(missing_ok=True)
+
+
 @dataclass
 class Session:
     """A Claude Code session."""
@@ -55,6 +129,8 @@ class Session:
     guest_code: str = ""
     guest_name: str = ""
     permission_level: str = "full"  # "full" or "read_only"
+    # Chat history
+    chat_session_id: str = ""  # Persistent chat session ID for resuming
 
     def __post_init__(self):
         self.working_dir = Path(self.working_dir)
@@ -98,6 +174,11 @@ class DarkCodeServer:
             db_path=self.config.config_dir / "guests.db"
         )
 
+        # Chat history persistence
+        self._chat_history = ChatHistory(
+            sessions_dir=self.config.config_dir / "chat_sessions"
+        )
+
         # Device binding state
         self._state = ServerState.AWAKE
         self._bound_device_id: Optional[str] = None
@@ -534,13 +615,19 @@ class DarkCodeServer:
 
         def read_line():
             if session.process and session.process.stdout:
-                return session.process.stdout.readline()
+                try:
+                    return session.process.stdout.readline()
+                except (IOError, OSError):
+                    return ""
             return ""
 
         while session.process and session.process.poll() is None:
             try:
                 line = await loop.run_in_executor(None, read_line)
                 if not line:
+                    # Check if process died
+                    if session.process and session.process.poll() is not None:
+                        break
                     await asyncio.sleep(0.01)
                     continue
 
@@ -560,18 +647,53 @@ class DarkCodeServer:
             except Exception:
                 break
 
+        # Process has exited - notify client and offer recovery
+        exit_code = session.process.poll() if session.process else -1
+        session.is_processing = False
+
+        # Read any stderr output
+        stderr_output = ""
+        if session.process and session.process.stderr:
+            try:
+                stderr_output = session.process.stderr.read() or ""
+            except Exception:
+                pass
+
+        # Send error to client with recovery action
+        try:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Claude process exited (code {exit_code})" + (f": {stderr_output[:200]}" if stderr_output else ""),
+                "recoverable": True,
+                "action": "reconnect",
+            }))
+            await session.websocket.send(json.dumps({
+                "type": "status",
+                "status": "closed",
+            }))
+        except Exception:
+            pass  # WebSocket might already be closed
+
     async def _handle_claude_output(self, session: Session, msg: dict):
         """Handle parsed output from Claude."""
         msg_type = msg.get("type", "")
 
         if msg_type == "assistant":
+            content = msg.get("message", {}).get("content") or msg.get("content")
             await session.websocket.send(json.dumps({
                 "type": "claude_message",
                 "role": "assistant",
-                "content": msg.get("message", {}).get("content") or msg.get("content"),
+                "content": content,
                 "timestamp": int(time.time() * 1000),
             }))
 
+            # Save assistant message to chat history
+            if session.chat_session_id and content:
+                self._chat_history.save_message(session.chat_session_id, {
+                    "role": "assistant",
+                    "content": content,
+                })
+
         elif msg_type == "tool_use":
             await session.websocket.send(json.dumps({
                 "type": "tool_use",
@@ -603,6 +725,51 @@ class DarkCodeServer:
                 "parsed": msg,
             }))
 
+    def _is_process_alive(self, session: Session) -> bool:
+        """Check if the Claude process is still running."""
+        if not session.process:
+            return False
+        return session.process.poll() is None
+
+    async def _write_to_process(self, session: Session, text: str) -> bool:
+        """Write text to Claude process stdin with error handling.
+
+        Returns True if write succeeded, False otherwise.
+        """
+        if not session.process or not session.process.stdin:
+            return False
+
+        # Check if process is still alive
+        if not self._is_process_alive(session):
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": "Claude process has terminated. Please reconnect.",
+                "recoverable": True,
+                "action": "reconnect",
+            }))
+            return False
+
+        try:
+            session.process.stdin.write(text + "\n")
+            session.process.stdin.flush()
+            return True
+        except BrokenPipeError:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": "Claude process closed unexpectedly. Please reconnect.",
+                "recoverable": True,
+                "action": "reconnect",
+            }))
+            return False
+        except (IOError, OSError) as e:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Failed to communicate with Claude: {e}",
+                "recoverable": True,
+                "action": "reconnect",
+            }))
+            return False
+
     async def _handle_message(self, session: Session, msg: dict):
         """Handle an authenticated message.
 
@@ -631,21 +798,31 @@ class DarkCodeServer:
                 }))
                 return
 
-            if text and session.process and session.process.stdin:
-                session.is_processing = True
-                session.last_active = time.time()
-                session.message_count += 1
-                # Text goes to Claude's stdin - Claude handles its own sandboxing
-                session.process.stdin.write(text + "\n")
-                session.process.stdin.flush()
-                await session.websocket.send(json.dumps({
-                    "type": "status",
-                    "status": "processing",
-                }))
+            if text:
+                # Check process health and write
+                if await self._write_to_process(session, text):
+                    session.is_processing = True
+                    session.last_active = time.time()
+                    session.message_count += 1
+
+                    # Save user message to chat history
+                    if session.chat_session_id:
+                        self._chat_history.save_message(session.chat_session_id, {
+                            "role": "user",
+                            "content": text,
+                        })
+
+                    await session.websocket.send(json.dumps({
+                        "type": "status",
+                        "status": "processing",
+                    }))
 
         elif msg_type == "abort":
-            if session.process:
-                session.process.send_signal(subprocess.signal.SIGINT)
+            if session.process and self._is_process_alive(session):
+                try:
+                    session.process.send_signal(subprocess.signal.SIGINT)
+                except (ProcessLookupError, OSError):
+                    pass  # Process already dead
                 session.is_processing = False
                 await session.websocket.send(json.dumps({
                     "type": "status",
@@ -653,22 +830,329 @@ class DarkCodeServer:
                 }))
 
         elif msg_type == "accept_edit":
-            if session.process and session.process.stdin:
-                session.process.stdin.write("y\n")
-                session.process.stdin.flush()
+            await self._write_to_process(session, "y")
 
         elif msg_type == "reject_edit":
-            if session.process and session.process.stdin:
-                session.process.stdin.write("n\n")
-                session.process.stdin.flush()
+            await self._write_to_process(session, "n")
 
         elif msg_type == "get_session_info":
             await session.websocket.send(json.dumps({
                 "type": "session_info",
                 "sessionId": session.id,
+                "chatSessionId": session.chat_session_id,
                 "workingDir": str(session.working_dir),
                 "isProcessing": session.is_processing,
                 "messageCount": session.message_count,
+                "processAlive": self._is_process_alive(session),
+            }))
+
+        elif msg_type == "set_chat_session":
+            # Set or create a chat session ID for persistence
+            chat_session_id = msg.get("chatSessionId", "")
+            if chat_session_id:
+                session.chat_session_id = chat_session_id
+            else:
+                # Generate a new one based on working dir and time
+                session.chat_session_id = hashlib.sha256(
+                    f"{session.working_dir}-{time.time()}".encode()
+                ).hexdigest()[:16]
+
+            await session.websocket.send(json.dumps({
+                "type": "chat_session_set",
+                "chatSessionId": session.chat_session_id,
+            }))
+
+        elif msg_type == "get_chat_history":
+            # Return chat history for a session
+            chat_session_id = msg.get("chatSessionId", session.chat_session_id)
+            if chat_session_id:
+                messages = self._chat_history.load(chat_session_id)
+                await session.websocket.send(json.dumps({
+                    "type": "chat_history",
+                    "chatSessionId": chat_session_id,
+                    "messages": messages,
+                }))
+            else:
+                await session.websocket.send(json.dumps({
+                    "type": "chat_history",
+                    "messages": [],
+                }))
+
+        elif msg_type == "list_chat_sessions":
+            # List all saved chat sessions
+            sessions = self._chat_history.list_sessions()
+            await session.websocket.send(json.dumps({
+                "type": "chat_sessions_list",
+                "sessions": sessions,
+            }))
+
+        elif msg_type == "delete_chat_session":
+            # Delete a chat session
+            chat_session_id = msg.get("chatSessionId", "")
+            if chat_session_id:
+                self._chat_history.delete(chat_session_id)
+                await session.websocket.send(json.dumps({
+                    "type": "chat_session_deleted",
+                    "chatSessionId": chat_session_id,
+                }))
+
+        # === File Operations ===
+        elif msg_type == "list_files":
+            # List files in a directory
+            path = msg.get("path", "")
+            await self._handle_list_files(session, path)
+
+        elif msg_type == "create_directory":
+            # Create a new directory
+            path = msg.get("path", "")
+            await self._handle_create_directory(session, path)
+
+        elif msg_type == "read_file":
+            # Read file contents
+            path = msg.get("path", "")
+            await self._handle_read_file(session, path)
+
+        elif msg_type == "delete_file":
+            # Delete a file or directory
+            path = msg.get("path", "")
+            await self._handle_delete_file(session, path)
+
+    def _resolve_safe_path(self, session: Session, path: str) -> Optional[Path]:
+        """Resolve a path safely within the working directory.
+
+        Returns None if path would escape the working directory.
+        """
+        if not path:
+            return session.working_dir
+
+        # Resolve the path relative to working directory
+        try:
+            target = (session.working_dir / path).resolve()
+            # Ensure it's within working directory (prevent directory traversal)
+            if session.working_dir.resolve() in target.parents or target == session.working_dir.resolve():
+                return target
+            # Also allow if target starts with working_dir (for paths inside)
+            working_resolved = session.working_dir.resolve()
+            if str(target).startswith(str(working_resolved)):
+                return target
+        except (ValueError, OSError):
+            pass
+        return None
+
+    async def _handle_list_files(self, session: Session, path: str):
+        """List files in a directory."""
+        target = self._resolve_safe_path(session, path)
+
+        if not target:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": "Invalid path: outside working directory",
+            }))
+            return
+
+        if not target.exists():
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Path does not exist: {path}",
+            }))
+            return
+
+        if not target.is_dir():
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Not a directory: {path}",
+            }))
+            return
+
+        try:
+            files = []
+            for item in sorted(target.iterdir()):
+                stat = item.stat()
+                files.append({
+                    "name": item.name,
+                    "path": str(item.relative_to(session.working_dir)),
+                    "isDirectory": item.is_dir(),
+                    "size": stat.st_size if item.is_file() else 0,
+                    "modified": int(stat.st_mtime * 1000),
+                })
+
+            await session.websocket.send(json.dumps({
+                "type": "file_list",
+                "path": path or ".",
+                "files": files,
+            }))
+        except PermissionError:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Permission denied: {path}",
+            }))
+        except Exception as e:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Failed to list files: {e}",
+            }))
+
+    async def _handle_create_directory(self, session: Session, path: str):
+        """Create a new directory."""
+        if not path:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": "Path is required",
+            }))
+            return
+
+        target = self._resolve_safe_path(session, path)
+
+        if not target:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": "Invalid path: outside working directory",
+            }))
+            return
+
+        if target.exists():
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Path already exists: {path}",
+            }))
+            return
+
+        try:
+            target.mkdir(parents=True, exist_ok=False)
+            await session.websocket.send(json.dumps({
+                "type": "directory_created",
+                "path": path,
+            }))
+        except PermissionError:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Permission denied: {path}",
+            }))
+        except Exception as e:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Failed to create directory: {e}",
+            }))
+
+    async def _handle_read_file(self, session: Session, path: str):
+        """Read file contents."""
+        if not path:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": "Path is required",
+            }))
+            return
+
+        target = self._resolve_safe_path(session, path)
+
+        if not target:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": "Invalid path: outside working directory",
+            }))
+            return
+
+        if not target.exists():
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"File not found: {path}",
+            }))
+            return
+
+        if not target.is_file():
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Not a file: {path}",
+            }))
+            return
+
+        # Limit file size to 1MB for reading
+        MAX_READ_SIZE = 1024 * 1024
+        if target.stat().st_size > MAX_READ_SIZE:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"File too large (max {MAX_READ_SIZE // 1024}KB)",
+            }))
+            return
+
+        try:
+            content = target.read_text(encoding="utf-8", errors="replace")
+            await session.websocket.send(json.dumps({
+                "type": "file_content",
+                "path": path,
+                "content": content,
+                "size": len(content),
+            }))
+        except PermissionError:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Permission denied: {path}",
+            }))
+        except Exception as e:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Failed to read file: {e}",
+            }))
+
+    async def _handle_delete_file(self, session: Session, path: str):
+        """Delete a file or empty directory."""
+        if not path:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": "Path is required",
+            }))
+            return
+
+        target = self._resolve_safe_path(session, path)
+
+        if not target:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": "Invalid path: outside working directory",
+            }))
+            return
+
+        if not target.exists():
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Path not found: {path}",
+            }))
+            return
+
+        # Don't allow deleting the working directory itself
+        if target.resolve() == session.working_dir.resolve():
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": "Cannot delete working directory",
+            }))
+            return
+
+        try:
+            if target.is_file():
+                target.unlink()
+            elif target.is_dir():
+                # Only delete empty directories for safety
+                if any(target.iterdir()):
+                    await session.websocket.send(json.dumps({
+                        "type": "error",
+                        "message": "Directory not empty",
+                    }))
+                    return
+                target.rmdir()
+
+            await session.websocket.send(json.dumps({
+                "type": "file_deleted",
+                "path": path,
+            }))
+        except PermissionError:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Permission denied: {path}",
+            }))
+        except Exception as e:
+            await session.websocket.send(json.dumps({
+                "type": "error",
+                "message": f"Failed to delete: {e}",
             }))
 
     async def _destroy_session(self, session: Session):
